%
%
%  Spots Create Channel Function for Imaris 7.3.0
%
%  Copyright Bitplane AG 2011
%
%
%  Installation:
%  - Copy this file into the XTensions folder in the Imaris installation directory.
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%       <Submenu name="Spots Functions">
%        <Item name="Spots to Spots Closest Distance" icon="Matlab">
%          <Command>MatlabXT::XTSpotsClosestDistance(%i)</Command>
%        </Item>
%       </Submenu>
%      </Menu>
%      <SurpassTab>
%        <SurpassComponent name="bpSpots">
%          <Item name="Spots to Spots Closest Distance" icon="Matlab">
%            <Command>MatlabXT::XTSpotsClosestDistance(%i)</Command>
%          </Item>
%        </SurpassComponent>
%      </SurpassTab>
%    </CustomTools>
% 
% !!! Additional informations:
% A new channel is added to the dataset
% Original dataset size is maintained
%

function XTSpotsClosestDistance(aImarisApplicationID)

% connect to Imaris interface
if ~isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
  javaaddpath ImarisLib.jar
  vImarisLib = ImarisLib;
  if ischar(aImarisApplicationID)
    aImarisApplicationID = round(str2double(aImarisApplicationID));
  end
  vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
else
  vImarisApplication = aImarisApplicationID;
end

% the user has to create a scene with some spots
vSurpassScene = vImarisApplication.GetSurpassScene;
if isequal(vSurpassScene, [])
    msgbox('Please create some Spots in the Surpass scene!');
    return;
end

% get the spots
vSpots = vImarisApplication.GetFactory.ToSpots(vImarisApplication.GetSurpassSelection);

% search the spots if not previously selected
if ~vImarisApplication.GetFactory.IsSpots(vSpots)
    for vChildIndex = 1:vSurpassScene.GetNumberOfChildren
        vDataItem = vSurpassScene.GetChild(vChildIndex - 1);
        if vImarisApplication.GetFactory.IsSpots(vDataItem)
            vSpots = vImarisApplication.GetFactory.ToSpots(vDataItem);
            break;
        end
    end
    % did we find the spots?
    if isempty(vSpots)
        msgbox('Please create some spots!');
        return;
    end
end

[vAnswer, vOk] = listdlg('ListString',{'Spots Statistics','Spots Radius','DataSet Channel'}, ...
    'InitialValue', 1, 'SelectionMode','multiple', ...
    'ListSize',[300 100], 'Name','Spots to Spots Closest Distance', ...
    'PromptString',{'Please select the result mode:'});
if vOk<1, return; end
vMode = [0, 0, 0];
vMode(vAnswer) = 1;

% get the spots coordinates
vSpotsXYZ = vSpots.GetPositionsXYZ;
vSpotsTime = vSpots.GetIndicesT;
vSpotsRadius = vSpots.GetRadiiXYZ;

vDistBorder = false;
% The Border distance calculation is only available for spherical spots, and not ellipsoids
if(isequal(vSpotsRadius(:,1),vSpotsRadius(:,2),vSpotsRadius(:,3)))
	vAnswer = questdlg('Please select the distance mode:', 'Spots to Spots Closest Distance', ...
		'Center', 'Border','Cancel', 'Center');
	if(isequal(vAnswer, 'Cancel') || isempty(vAnswer)), return, end
	vDistBorder = isequal(vAnswer, 'Border');
end

vMinDists = zeros(1, numel(vSpotsTime));
if vMode(1)
  vMeanDists = vMinDists;
  vMaxDists = vMinDists;
end

if vMode(3)
  % add vChannel to current dataset
  vImarisApplication.DataSetPushUndo('Spots to Spots Closest Distance');
  [vDataSet, vData, vMin, vMax] = InitNewChannel(vImarisApplication);
  vMaxTime = vDataSet.GetSizeT - 1;
  vChannel = vDataSet.GetSizeC - 1;
else
  vMaxTime = max(vSpotsTime);
end

vProgressDisplay = waitbar(0,'Creating DataSet');
for vTime = 0:vMaxTime
  if vMode(3) && vTime > 0
    vData(:) = 0;
  end
  
  vSpotsT = vSpotsTime == vTime;
  vSpotsIndex = find(vSpotsT');
  for vIndex = vSpotsIndex
    vOthers = vSpotsT;
    vOthers(vIndex) = false;
    vX = vSpotsXYZ(vOthers, 1) - vSpotsXYZ(vIndex, 1);
    vY = vSpotsXYZ(vOthers, 2) - vSpotsXYZ(vIndex, 2);
    vZ = vSpotsXYZ(vOthers, 3) - vSpotsXYZ(vIndex, 3);
    vDist = sqrt(sum( ([vX, vY, vZ]).^2, 2 ));
    
    if vDistBorder
      vDist = max(vDist - vSpotsRadius(vOthers, 1) - vSpotsRadius(vIndex, 1), 0);
    end
    vMinDist = min(vDist);
    vMinDists(vIndex) = vMinDist;
    
    if vMode(1)
      vMeanDists(vIndex) = mean(vDist);
      vMaxDists(vIndex) = max(vDist);
    end

    if vMode(3)
      vPos = vSpotsXYZ(vIndex, :);
      vRadius = vSpotsRadius(vIndex, :);
      vData = DrawSphere(vData, vPos, vRadius, vMin, vMax, vMinDist);
    end
  end
  if vMode(3) && ~isempty(vSpotsIndex)
    if strcmp(vDataSet.GetType, 'eTypeUInt8')
      vDataSet.SetDataVolumeBytes(vData, vChannel, vTime);
    elseif strcmp(vDataSet.GetType, 'eTypeUInt16')
      vDataSet.SetDataVolumeShorts(vData, vChannel, vTime);
    else
      vDataSet.SetDataVolumeFloats(vData, vChannel, vTime);
    end
  end
  
  waitbar(vTime/vMaxTime)
end
close(vProgressDisplay);

vRGBA = vSpots.GetColorRGBA;
if vMode(3)
  vDataSet.SetChannelColorRGBA(vChannel, vRGBA);
  vDataSet.SetChannelName(vChannel, [char(vSpots.GetName), 'ClosestDist']);
  vImarisApplication.SetDataSet(vDataSet);
end

if vMode(1) || vMode(2)
  vNewSpots = vImarisApplication.GetFactory.CreateSpots;
  if vMode(2)
    vNewSpots.Set(vSpotsXYZ, vSpotsTime, vMinDists);
  else
    vNewSpots.Set(vSpotsXYZ, vSpotsTime, vSpotsRadius(:,1));
    vNewSpots.SetRadiiXYZ(vSpotsRadius);
  end
  vNewSpots.SetColorRGBA(vRGBA);
  vNewSpots.SetName([char(vSpots.GetName), 'Dist']);
  if vMode(1)
    vInd = 1:numel(vMinDists);
    vIds = vInd - 1;
    vUnits(vInd) = { char(vImarisApplication.GetDataSet.GetUnit) };
    vFactors(vInd) = {'Spot'};
    vFactorNames = {'Category'};
    vNames(vInd) = {'DistMin'};
    vNewSpots.AddStatistics(vNames, vMinDists, vUnits, vFactors, vFactorNames, vIds);
    vNames(vInd) = {'DistMean'};
    vNewSpots.AddStatistics(vNames, vMeanDists, vUnits, vFactors, vFactorNames, vIds);
    vNames(vInd) = {'DistMax'};
    vNewSpots.AddStatistics(vNames, vMaxDists, vUnits, vFactors, vFactorNames, vIds);
  end  
  vSpots.SetVisible(0);
  vSpots.GetParent.AddChild(vNewSpots, -1);
end


function aData = DrawSphere(aData, aPos, aRad, aMin, aMax, aValue)
vSize = [size(aData, 1), size(aData, 2), size(aData, 3)];
aPos = (aPos - aMin) ./ (aMax - aMin) .* vSize + 0.5;
aRad = aRad ./ (aMax - aMin) .* vSize;
vPosMin = round(max(aPos - aRad, 1));
vPosMax = round(min(aPos + aRad, vSize));
vPosX = vPosMin(1):vPosMax(1);
vPosY = vPosMin(2):vPosMax(2);
vPosZ = vPosMin(3):vPosMax(3);
[vX, vY, vZ] = ndgrid(vPosX, vPosY, vPosZ);
vDist = ((vX - aPos(1))/aRad(1)).^2 + ((vY - aPos(2))/aRad(2)).^2 + ...
     ((vZ - aPos(3))/aRad(3)).^2;
vInside = vDist < 1;
vCube = aData(vPosX, vPosY, vPosZ);
vCube(vInside) = aValue;
aData(vPosX, vPosY, vPosZ) = vCube;


function [aDataSet, aData, aMin, aMax] = InitNewChannel(aImaris)
aDataSet = aImaris.GetDataSet.Clone;
aMin = [aDataSet.GetExtendMinX, aDataSet.GetExtendMinY, aDataSet.GetExtendMinZ];
aMax = [aDataSet.GetExtendMaxX, aDataSet.GetExtendMaxY, aDataSet.GetExtendMaxZ];
vSize = [aDataSet.GetSizeX, aDataSet.GetSizeY, aDataSet.GetSizeZ];
if strcmp(aDataSet.GetType, 'eTypeUInt8')
  aData = zeros(vSize, 'uint8');
elseif strcmp(aDataSet.GetType, 'eTypeUInt16')
  aData = zeros(vSize, 'uint16');
else
  aData = zeros(vSize, 'single');
end
aDataSet.SetSizeC(aDataSet.GetSizeC + 1);
